//Todo: header and stuff
//Copyright 2004,2005 Rob Reilink

#include "sd.h"
#include "exi.h"
#include <linux/string.h>
#include "block.h"
#include "fat.h"
#include "file.h"

//Todo: place this in disk.c or something
#include <memlayout.h>
#include <elf.h>
#include <types.h>

/* compatibility with linux kernel driver (ugly)*/
#define printk printf
#define SUCCESS 0
#define FAIL -1

#define GCN_SD_SPEED 4
#define PFX
#define KERN_INFO
#define KERN_ALERT
#define KERN_ERR

/* COPY-PASTED FROM THE LINUX DRIVER */
/*
 *  EXI/SD functions
 *
 */
static char gcn_sd_cmd_short_nofinish(unsigned char cmd) {
	unsigned char dat;
	int tries;
	exi_select(1,0,GCN_SD_SPEED);
	dat=cmd|0x40;
	exi_write(1,&dat,1);
	dat=255;
	for (tries=0;(tries < 30) && (dat == 0xff);tries++)
		exi_read(1,&dat,1);

	return dat;
}

static char gcn_sd_cmd_nofinish(unsigned char cmd, unsigned long data) {
	unsigned char snd[6],ret;
	int tries;
	exi_select(1,0,GCN_SD_SPEED);
	snd[0]=cmd|0x40;
	snd[1]=data >> 24;
	snd[2]=(data >> 16) & 0xff;
	snd[3]=(data >> 8) & 0xff;
	snd[4]=data & 0xff;
	snd[5]=0x95;

	exi_write(1,&snd,6);
	
	//wait loop
	ret=0xff;
	for(tries=0;(tries<100) && (ret==0xff);tries++)
		exi_read(1,&ret,1);
	return ret;
}

static void gcn_sd_finish(void) {
	unsigned char zero=0;
	exi_deselect(1);
	/*last dummy write. This may look weird, accessing the device
	  after it has been deselected, but the SD manual states that a SD
	  card needs 8 extra clock cycles *after* deselection, and this is
	  how to generate those
	*/  
	exi_write(1, &zero, 1);

}

inline static char gcn_sd_cmd(unsigned char cmd) {
	char i=gcn_sd_cmd_nofinish(cmd,0);
	gcn_sd_finish();
	return i;
} 

inline static char gcn_sd_cmd_short(unsigned char cmd) {
	char i=gcn_sd_cmd_short_nofinish(cmd);
	gcn_sd_finish();
	return i;
}

/* returns: found?SUCCESS:FAIL */
static int gcn_sd_init(void) {
	int i,ret; long c;
	int tries=0;

	//Send 80 clocks while not selected
	exi_select(1,1,GCN_SD_SPEED);  //SLOOW TODO:fast, select=-1 ofzo (niets selecten)
	exi_deselect(1);
	for(i=0;i<20;i++) {
	 	c=0xffffffff;
		exi_write(1,&c,4);
	}

	
	//send CMD0 to init
	gcn_sd_cmd(0);
	gcn_sd_cmd_short(0);
	for(tries=0;tries<100;tries++) {
		if ((ret=gcn_sd_cmd_short(58))==0) return SUCCESS;
		printk(KERN_ALERT PFX "cmd 58 returned %d\n",ret);
		if ((ret=gcn_sd_cmd_short(1))==0) return SUCCESS;
		printk(KERN_ALERT PFX "cmd 1 returned %d\n",ret);
	}
	return FAIL;
}

static int gcn_sd_read_csd(void) {
	unsigned char c;
	unsigned char data[16];
	int tries;
	c=gcn_sd_cmd_short_nofinish(9);
	if (c!=0) {
		gcn_sd_finish();
		printk(KERN_INFO PFX "could not read csd(%d)\n",c);
		return FAIL;
	}
	for (tries=0;(tries<100000) & (c!=0xfe);tries++)
		exi_read(1,&c,1);

	if (c!=0xfe) {
		gcn_sd_finish();
		printk(KERN_ERR PFX "error: no start token after csd read request\n");
		return FAIL;		
	}
	exi_read(1,data,16);
	gcn_sd_finish();
	int i;
	printk(KERN_INFO PFX "CSD data:");
	for (i=0;i<16;i++) printk ("%x ",data[i]);
	printk("\n");

	return SUCCESS;
}

static int gcn_sd_block_read(unsigned long sector,void * data) {
	unsigned char c;
	long dummy;
	int tries=10;
	int tokentries;
retry:
	if ((tries--)<=0) {
		printk(KERN_ERR PFX "error: reading failed\n");
		return FAIL;
	}
	c=gcn_sd_cmd_nofinish(17,(sector) << 9);

	if (c!=0) {
		gcn_sd_finish();
		printk(KERN_ERR PFX "error: read command returned %d while reading sector %lu\n" ,c,sector);
		goto retry;
	}

	//Wait for start of data token 0xfe TODO put this in a function
	for (tokentries=0;(tokentries<100000) & (c!=0xfe);tokentries++)
		exi_read(1,&c,1);

	if (c!=0xfe) {
		gcn_sd_finish();
		printk(KERN_ERR PFX "error: no start token after read request sector=%lu\n" ,sector);
		goto retry;
	}
	exi_read(1,data,512);
	exi_read(1,&dummy,3); //response,CRC, dummy. todo: use it?

	gcn_sd_finish();

	return SUCCESS;
	
}

void dump_sector(char *data)
{
	int i;
	for (i=0;i<512;i++) {
		printf("%02x ",data[i]);
		if ((i % 16)==0) printf("\n");
	}
}

int sd_block_read(int dev, int start, int size, void *buffer) {
	int res;
	while (size)
	{
		res = gcn_sd_block_read(start,buffer);
		if (res < 0)
			return res;
		buffer = ((unsigned char*)buffer) + 512;
		start++;
		size--;
	}
	return 0;

}
struct partition_s
{
	unsigned char state, start_head;
	unsigned short start_cyl;
	unsigned char type, end_head;
	unsigned short end_cyl;
	unsigned long start_lba, size;
};
//Todo: Yak. Put this in a file independent of device!

int sd_read_parttable(struct partition_s *part)
{
	unsigned char sector[512];
	if (gcn_sd_block_read(0,sector))
		return -1;
	memcpy(part, sector + 446, 64);
	return 0;
}

inline unsigned short swab16(unsigned short l)
{
	return (l >> 8) | (l<<8);
}

inline unsigned long swab32(unsigned long l)
{
	return ((l >> 24) & 0xFF) | ((l >> 8) & 0xFF00) | ((l << 8) & 0xFF0000) | ((l<<24) & 0xFF000000);
}

//Returns on failiure, runs code on success
static block_dev_desc_t bdev;
void sd_boot(void)
{
	printf("TEST A\n");
	if (gcn_sd_init()==FAIL) return;

	printf("SD found, booting...\n");
	//gcn_sd_block_read(0,data);
	struct partition_s parttable[4];
	if (sd_read_parttable(parttable))
		return;
	
	int offset = (int)swab32(parttable[0].start_lba);
	int size = (int)swab32(parttable[0].size);
	
	//TODO: total if ((!offset) || (!size) || ((offset + size) > total))
	if ((!offset) || (!size))
	{
		printf("first partition is invalid!\n");
		return;
	}
		printf("first partition is found!\n");
//	for(;;);
	bdev.block_read = sd_block_read;
	bdev.dev = 1;
	fat_register_device(&bdev, offset);
	
	if (file_fat_detectfs() < 0)
	{
		printf("no fat detected.\n");
		return;
	} 
	
	file_fat_ls("/");
	//(check:bytes or sectors?)
	file_detectfs();
	if (file_read("/main.elf", ELF_FILE_BUFFER, 8*1024*1024) < 0) {
		printf("file read error\n");
		return;
	}
	printf("done reading\n");

	//TODO: this somewhere else
	void (*entrypoint)();
	entrypoint = (void(*)(void))load_elf_image((unsigned long)ELF_FILE_BUFFER);
	
	if (entrypoint != NULL) {
		// clear interrupt mask
		*(volatile unsigned long*)0xCC003004 = 0;

		exception_close();
		printf("starting execution at %p\n", entrypoint);
		entrypoint();
	}


}
