#include "httpd/httpd.h"
#include <linux/string.h>
#include "lwip/debug.h"
#include "lwip/stats.h"
#include "lwip/tcp.h"

#define FLASH_IMAGE ((unsigned char*)0x80800000)

struct response_flash_put_priv_s
{
	int content_length;
	int content_type_ok;
	int flash_addr, flash_len;
	int hdr_state;
	
	unsigned char* dst;
	
	int upload_ptr;
	
#define FS_UPLOAD    0
#define FS_PREPARE   1
#define FS_FLASHING  2
#define FS_DONE      3
	
	int flash_state;
	
	int wantcrc;
	
	int marker_pos;
	
		/* POST stuff */
	char boundary[128];
	int boundary_ptr;
	int mime_state; // 0: wait for boundary, 1: header, 2: data

	int html_bytes_sent;
};

int response_flash_put_process_request(struct http_state *http, const char *method, const char *url)
{
	if (strcmp(method, "POST"))
		return 0;
		
	if (strcmp(url, "/FLASH"))
		return 0;

	http->response_priv = mem_malloc(sizeof(struct response_flash_put_priv_s));
	if (!http->response_priv)
		return 0;
	struct response_flash_put_priv_s *priv = http->response_priv;
	
	memset(http->response_priv, 0, sizeof(struct response_flash_put_priv_s));

	priv->flash_state = FS_UPLOAD;
	priv->hdr_state = 0;
	priv->dst = (unsigned char*)FLASH_IMAGE;
	http->code = 400;
	return 1;
}

void response_flash_put_process_header(struct http_state *http, const char *option, const char *val)
{
	struct response_flash_put_priv_s *priv = http->response_priv;
	
	if (!strcmp(option, "Content-Type"))
		if (!strncmp(val, "multipart/form-data", 19))
		{
			char *boundary = strstr(val, "boundary=");
			if (boundary)
			{
				strcpy(priv->boundary, "\r\n--");
				strcat(priv->boundary, boundary + 9);
				priv->boundary_ptr = 2;
				priv->mime_state = 0;
				http->code = 200;
			}
		}
}

void response_flash_put_start_response(struct http_state *http)
{
}

int response_flash_put_do_header(struct http_state *http)
{
	struct response_flash_put_priv_s *priv = http->response_priv;
	
	const char *t=0, *o=0;
	switch (priv->hdr_state)
	{
	case 0:
		t = "Content-Type";
		o = "text/html";
		break;
	case 1:
		return httpd_do_std_header(http);
	}
	
	int av = httpd_available_sendbuffer(http);
	if (av < (strlen(t) + strlen(o) + 4))
		return 1;
	
	httpd_put_sendbuffer_string(http, t);
	httpd_put_sendbuffer_string(http, ": ");
	httpd_put_sendbuffer_string(http, o);
	httpd_put_sendbuffer_string(http, "\r\n");
	++priv->hdr_state;
	return 2;
}

static void flash_show_response(struct http_state *http, int type, int nr, const char *message)
{
//	struct response_flash_put_priv_s *priv = http->response_priv;
	if (type == 1)
	{
		httpd_put_sendbuffer_string(http, "<script>setAction('");
		httpd_put_sendbuffer_string(http, message);
		httpd_put_sendbuffer_string(http, "');</script>\n");
	} else
	{
		httpd_put_sendbuffer_string(http, "<p>");
		httpd_put_sendbuffer_string(http, message);
		httpd_put_sendbuffer_string(http, "</p>");
	}
}

int response_flash_put_do_data(struct http_state *http)
{
	struct response_flash_put_priv_s *priv = http->response_priv;
	
	int av = httpd_available_sendbuffer(http);

	if (av < 256)
		return 1;
		
	char resp[256];
	
	if (http->code != 200)
		priv->flash_state = FS_DONE;
	
	switch (priv->flash_state)
	{
		case FS_UPLOAD:
			if (priv->html_bytes_sent != -1)
			{
				const char *t = 
"<html><body><table align=\"center\"><tr><td><div style=\"font-size:8pt;padding:2px;border:solid black 1px\">\n"
"<span id=\"action\">blub</span>\n"
"<span id=\"progress0\">&nbsp; &nbsp;</span>\n"
"<span id=\"progress1\">&nbsp; &nbsp;</span>\n"
"<span id=\"progress2\">&nbsp; &nbsp;</span>\n"
"<span id=\"progress3\">&nbsp; &nbsp;</span>\n"
"<span id=\"progress4\">&nbsp; &nbsp;</span>\n"
"<span id=\"progress5\">&nbsp; &nbsp;</span>\n"
"<span id=\"progress6\">&nbsp; &nbsp;</span>\n"
"<span id=\"progress7\">&nbsp; &nbsp;</span>\n"
"<span id=\"val\"></span></div></td></tr></table><script language=\"javascript\">\n"
"/*  tFix: Autor 2003 Kristof Lipfert, Duesseldorf                 */\n"
"function tFix(wert,ds){\n"
"        var wert=(wert.toFixed)?wert.toFixed(ds):Math.floor(wert)+\".\"+\n"
"        (Math.pow(10,ds)+Math.round((wert-Math.floor(wert))*Math.pow(10,ds))+\"\").substr(1,ds);\n"
"        return wert;\n"
"}\n"
"function progress_update(x) \n"
"{\n"
"	y = x * 8 / (32 * 1024 * 1024);\n"
"	p = x * 100 / (32 * 1024 * 1024);\n"
"	for (i=0; i<8; ++i)\n"
"		if (y >= i)\n"
"			document.getElementById('progress'+i).style.backgroundColor = 'blue';\n"
"		else\n"
"			document.getElementById('progress'+i).style.backgroundColor = 'white';\n"
"	document.getElementById('val').innerHTML = tFix(p, 2) + \"%\";\n"
"}\n"
"function setAction(x) { document.getElementById('action').innerHTML = x; } \n"
"setAction('uploading...'); \n"
"</script>\n";

				int tr = strlen(t + priv->html_bytes_sent);
				if (tr)
				{
					if (tr > av)
						tr = av;
					httpd_put_sendbuffer(http, t + priv->html_bytes_sent, tr);
					av -= tr;
					priv->html_bytes_sent += tr;
					if (av < 256)
						return 1;
				} else
					priv->html_bytes_sent = -1;
			}
			return 1;
		case FS_PREPARE:
		{
			if (priv->content_length > (1<<20))
			{
				flash_show_response(http, 2, 8, "IPL size > 1MB. This is not (yet) supported.");
				priv->flash_state = FS_DONE;
				return 1;
			}
			flash_show_response(http, 1, 5, "starting flashing...");
			priv->flash_state = FS_FLASHING;
			return 1;
		}
		case FS_FLASHING:
		{
			do_flash(1<<20, priv->dst, priv->content_length);
#if 0
				if (memcmp(sector, buffer, 528))
				{
					sprintf(resp, "verify failed at %08x", priv->i);
					flash_show_response(http, 2, 8, resp);
					priv->flash_state = FS_DONE;
					return 1;
				}
				
				priv->i += 512;
					/* sent less updates for html */
				if (!(priv->i &  0x1FF))
				{
					sprintf(resp, "<script>progress_update(%d);</script>\r\n", priv->i);
					httpd_put_sendbuffer_string(http, resp);
					return 1;
				}
			}
#endif
			flash_show_response(http, 1, 999, "finished.");
			priv->flash_state = FS_DONE;
			return 0;
		case FS_DONE:
			return 0;
		}
	}

	return 1;
}

int response_flash_put_process_data(struct http_state *http, void *data, int len)
{
	struct response_flash_put_priv_s *priv = http->response_priv;
	
	int consumed_len = len;

	if (priv->flash_state != FS_UPLOAD)
		return len;
		
	unsigned char *d = data;
	while (len)
	{
		if (*d == priv->boundary[priv->boundary_ptr])
		{
			++priv->boundary_ptr; ++d; --len;
			if (!priv->boundary[priv->boundary_ptr])
			{
				priv->boundary_ptr = 0;
				if (priv->mime_state == 1)
				{
					priv->content_length = priv->upload_ptr;
					
					int consumed = 0;
					while (consumed < priv->content_length)
					{
						if (!memcmp(priv->dst + consumed, "\n\r\n", 3))
						{
							consumed += 3;
							break;
						}
						if (!memcmp(priv->dst + consumed, "\n\n", 2))
						{
							consumed += 2;
							break;
						}
						++consumed;
					}
					
					priv->content_length -= consumed;
					priv->dst += consumed;
					
					priv->flash_state = FS_PREPARE;
				}
				++priv->mime_state;
			}
		} else
		{
			if (priv->mime_state == 1)
			{
				if (priv->boundary_ptr)
				{
					memcpy(priv->dst + priv->upload_ptr, priv->boundary, priv->boundary_ptr);
					priv->upload_ptr += priv->boundary_ptr;
					priv->boundary_ptr = 0;
				}
				priv->dst[priv->upload_ptr++] = *d++; --len;
			} else
			{
				++d; --len;
				priv->boundary_ptr = 0;
			}
		}
	}
	
	return consumed_len;
}

void response_flash_put_finish(struct http_state *http)
{
	struct response_flash_put_priv_s *priv = http->response_priv;
	mem_free(priv);
}
