#ifndef __irq_h
#define __irq_h

#include "processor.h"

void irq_ext(); // external IRQ occured - entry point from exception code
void irq_debug_ext(); // external IRQ occured - debug

typedef int irq_handler_proc(int irq, void *context);

int request_irq(int irq, irq_handler_proc *handler, void *context);
int request_debug_irq(int irq, irq_handler_proc *handler, void *context);
int free_irq(int irq, void *context);
int free_debug_irq(int irq, void *context);

#define IRQ_HSP       (18)
#define IRQ_DEBUG     (19)
#define IRQ_CP        (20)
#define IRQ_PE_FINISH (21)
#define IRQ_PE_TOKEN  (22)
#define IRQ_VI        (23)
#define IRQ_MEM       (24)
#define IRQ_DSP       (25)
#define IRQ_AI        (26)
#define IRQ_EXI       (27)
#define IRQ_SI        (28)
#define IRQ_DI        (29)
#define IRQ_RSW       (30)
#define IRQ_ERROR     (31)

#define cli(msr) do { msr = mfmsr(); mtmsr((msr) & ~0x8000); } while (0)
#define sti(msr) do { mtmsr(msr); } while (0)

typedef unsigned long mutex_t;
#define initialize_mutex(x) x = 0;
#define down(x) cli(x)
#define up(x) sti(x)

#endif
