#include <console.h>
#include <time.h>
#include <cache.h>

volatile long *dvd=(volatile long *)0xCC006000;

void dvd_reset(void)
{
	*(unsigned long*)0xcc006004 = 2;
	unsigned long v = *(unsigned long*)0xcc003024;
	*(unsigned long*)0xcc003024 = (v &~4) | 1;
	mdelay(1);
	*(unsigned long*)0xcc003024 = v | 5;
}

unsigned int dvd_read(void *dst, int len, unsigned int offset)
{
    unsigned char sector[1024+0x40];
    unsigned char* buffer = (unsigned char*)((unsigned int)sector & 0xffffffe0);
    dcache_inv(buffer, 1024);
    
    if (len > 1024)
        return 1;

    if (offset<0x57057C00) 
    {        
    	dvd[0] = 0x2E;
    	dvd[1] = 0;	
    	dvd[2] = 0xA8000000;
    	dvd[3] = offset >> 2;
    	dvd[4] = len+0x1f;
    	dvd[5] = (unsigned long)buffer;
    	dvd[6] = len+0x1f;
    	dvd[7] = 3; // enable reading!	
    	while (dvd[7] & 1);
    	memcpy(dst, buffer + (offset & 0x3), len);
    } else // Let's not read past end of DVD
    {
    	dvd[0] = 0x2E;
    	dvd[1] = 0;	
    	dvd[2] = 0xA8000000;
    	dvd[3] = 0x57057C00 >> 2;
    	dvd[4] = 0x400;
    	dvd[5] = (unsigned long)buffer;
    	dvd[6] = 0x400;
    	dvd[7] = 3; // enable reading!	
    	while (dvd[7] & 1);
    	memcpy(dst, buffer + (offset - 0x57057C00), len);        
    }

	if (dvd[0] & 0x4)
		return 1;
	return 0;
}

int dvd_seek(int offset)
{
	dvd[0] = 0x2E;
	dvd[1] = 0;
	dvd[2] = 0xAB000000;
	dvd[3] = offset >> 2;
	dvd[4] = 0;
	
	dvd[7] = 1;
	while (dvd[7] & 1);
	
	if (dvd[0] & 0x4)
		return 1;
	return 0;
}

unsigned int dvd_read_id(void *dst)
{
	if ((((int)dst) & 0xC0000000) == 0x80000000) // cached?
		dcache_inv(dst, 0x20);

	dvd[0] = 0x2E;
	dvd[1] = 0;
	
	dvd[2] = 0xA8000040;
	dvd[3] = 0;
	dvd[4] = 0x20;
	dvd[5] = (unsigned long)dst;
	dvd[6] = 0x20;
	dvd[7] = 3; // enable reading!

	while (dvd[7] & 1);

	if (dvd[0] & 0x4)
		return 1;
	return 0;
}

void dvd_motor_off()
{
	dvd[0] = 0x2E;
	dvd[1] = 0;
	
	dvd[2] = 0xe3000000;
	dvd[3] = 0;
	dvd[4] = 0;
	dvd[5] = 0;
	dvd[6] = 0;
	dvd[7] = 1; // enable reading!
	while (dvd[7] & 1);
}

unsigned int dvd_init()
{
//	dvd_reset();
	if (dvd_read_id(0x80000000))
		return -1;
	dcache_inv(0x80000000, 0x20);
	return 0;
}

void dvd_set_audio(int enable, int len)
{
	enable = !!enable;
	dvd[2] = 0xE4000000 | (enable<<16) | len;
	dvd[7] = 1;
	while (dvd[7] & 1);
}

void dvd_set_streaming(void)
{
	int streaming_enable = *(unsigned char*)0x80000008;
	dvd[1] = dvd[1];
	if (!streaming_enable)
		dvd_set_audio(0, 0);
	else
		dvd_set_audio(1, (*(unsigned char*)0x80000009) ? 0 : 0xA);

	while (dvd[7] & 1);
}

unsigned int dvd_get_error(void)
{
	dvd[2] = 0xE0000000;
	dvd[8] = 0;
	dvd[7] = 1;
	while (dvd[7] & 1);
	return dvd[8];
}

unsigned int dvd_get_diconfig(void)
{
	return dvd[9];
}

const char *dvd_get_error_message(void)
{
	switch (dvd_get_error()>>24)
	{
			/* as documented in NinDoc */
	case 0x00: return "Ready.";
	case 0x01: return "Cover is opened.";
	case 0x02: return "Disk change.";
	case 0x03: return "No Disk.";
	case 0x04: return "Motor stop.";
	case 0x05: return "Disk ID not read.";
	default: return "unknown";
	}
}


/**************** NPDP STUFF ***************/
/* by Dr.Matrix */

/* Note by tmbinc: I wasn't able to test this 
 stuff, as i don't have a NPDP. But Dr.Matrix 
 usually does things which work, i trust in him! */


unsigned int npdp_getid(unsigned char *dst)
{
	static unsigned char id[256] __attribute__((__aligned__(32)));
	dcache_inv(id, 0x40);

	dvd[0] = 0x2E;
	dvd[1] = 0;
	
	memcpy(dvd + 2, "NPDP-ID", 8);
	
	dvd[4] = 0x40;
	dvd[5] = (unsigned long)id;
	dvd[6] = 0x40;
	dvd[7] = 3; // enable reading!

	while (dvd[7] & 1);
	
	if (dvd[6])
		return 1;
	memcpy(dst, id, 0x40);
	return 0;
}

