/*
 * function to write to flash. Only works with modified IPL "ROM"s.
 */

#include <ipl.h>
#include <vsprintf.h>

int chip_erase(void)
{
	rom_write(0xAAA, 0xAA);
	rom_write(0x555, 0x55);
	rom_write(0xAAA, 0x80);
	rom_write(0xAAA, 0xAA);
	rom_write(0x555, 0x55);
	rom_write(0xAAA, 0x10);
	printf("erasing..\n");
	while (!(rom_read(0) & 0x80));
	return 0;
}

int erase_sector(int addr)
{
	rom_write(0xAAA, 0xAA);
	rom_write(0x555, 0x55);
	rom_write(0xAAA, 0x80);
	rom_write(0xAAA, 0xAA);
	rom_write(0x555, 0x55);
	rom_write(addr, 0x30);
	printf("\r%08x erase..\n", addr);
//	while (!(rom_read(addr) & 0x80)) ;
	while (rom_read(addr) != rom_read(addr));
	return 0;
}

int is_eraseblock(int addr)
{
	if (addr >= 0x10000)
		return !(addr & 0xFFFF);
	return ((addr == 0) || (addr == 0x4000) || (addr == 0x6000) || (addr == 0x8000));
}

int do_flash(int target, unsigned char *src, int len)
{
	ipl_set_config(0x0);
		// CFI query
	rom_write(0xAA, 0x98);

	printf("CFI data:\n");
	int addr;
	for (addr = 0; addr < 0x100; addr += 0x10)
	{
		int i;
		for (i=0; i<0x10; i += 2)
			printf("%02x ", rom_read(addr + i));

		for (i=0; i<0x10; i += 2)
		{
			int v = rom_read(addr + i);
			printf("%c", v >= 0x20 ? v : '.');
		}
		printf("\n");
	}
	
	printf("man id:\n");
	int i;
	for (i=0; i<10; ++i)
	{
		rom_write(0, 0xF0);
		rom_write(0xAAA, 0xAA); rom_write(0x555, 0x55); rom_write(0xAAA, 0x90);
		printf("%02x ", rom_read(i * 2));
	}
	printf("\n");
	
	rom_write(0xAAA, 0xAA); rom_write(0x555, 0x55); rom_write(0xAAA, 0x20);
	
	printf("erasing...\n");
	for (addr = target; addr < (target + len); ++addr)
		if (is_eraseblock(addr))
			erase_sector(addr);
	printf("flashing...\n");
	while (len)
	{
		rom_write(target, 0xA0);
		rom_write(target, *src);
		while (rom_read(target) != *src);
		++src;
		++target;
		--len;
	}
	printf("done!\n");
	return 0;
}
