/*
 * DOL loader over tcp/ip
 * Receives a DOL on port 4000 and executes it
 */

#include "lwip/debug.h"
#include "lwip/stats.h"
#include "lwip/tcp.h"
#include <console.h>
#include <types.h>
#include <processor.h>
#include <elf.h>

// ELF is read into a fixed buffer
//#define FILE_BUFFER ((u8*)0x80800000)
#include <memlayout.h>


#define TYPE_UNKNOWN 0
#define TYPE_DOL 1
#define TYPE_ELF 2

struct dolup_state {
	struct dol_s {
		u32 sec_pos[18];
		u32 sec_address[18];
		u32 sec_size[18];
		u32 bss_address, bss_size, entry_point;
		u32 dummy[7];
	} dolhdr;
	u32 filepos;
	u32 filesize;
	int type;
	int retries;
};
/*-----------------------------------------------------------------------------------*/
static void
conn_err(void *arg, err_t err)
{
	struct dolup_state *hs;
	
	hs = arg;
	mem_free(hs);
}
/*-----------------------------------------------------------------------------------*/
static void
close_conn(struct tcp_pcb *pcb, struct dolup_state *hs)
{
	void (*entrypoint)();
	
	printf("DOLUP: connection CLOSE\n");
	tcp_arg(pcb, NULL);
	tcp_sent(pcb, NULL);
	tcp_recv(pcb, NULL);
	tcp_close(pcb);
	
	entrypoint = NULL;
	switch (hs->type) {
		case TYPE_DOL:
			entrypoint = (void(*)())hs->dolhdr.entry_point;
			break;
		case TYPE_ELF:
			entrypoint = (void(*)(void))load_elf_image((unsigned long)ELF_FILE_BUFFER);
			break;
		default:
			printf("unknown file type\n");
	}
	mem_free(hs);
	
	if (entrypoint != NULL) {
		// clear interrupt mask
		*(volatile unsigned long*)0xCC003004 = 0;

		exception_close();
		printf("starting execution at %p\n", entrypoint);
		entrypoint();
	}
}
/*-----------------------------------------------------------------------------------*/
static err_t
dolup_poll(void *arg, struct tcp_pcb *pcb)
{
	struct dolup_state *hs;
	
	hs = arg;
	
 	// printf("Polll\n");
	if (hs == NULL) {
		printf("Null, close\n");
		tcp_abort(pcb);
		return ERR_ABRT;
	} else {
		++hs->retries;
		if (hs->retries == 40) {
			tcp_abort(pcb);
			return ERR_ABRT;
		}
//		send_data(pcb, hs);
	}
	
	return ERR_OK;
}
/*-----------------------------------------------------------------------------------*/
#if 0
static err_t
dolup_sent(void *arg, struct tcp_pcb *pcb, u16_t len)
{
	return ERR_OK;
}
#endif
/*-----------------------------------------------------------------------------------*/
static err_t
dolup_recv(void *arg, struct tcp_pcb *pcb, struct pbuf *p, err_t err)
{
	struct dolup_state *hs;
	
	hs = arg;
	
	if (err == ERR_OK && p != NULL) {
		// Inform TCP that we have taken the data.
		tcp_recved(pcb, p->tot_len);
		
		u8 * data = p->payload;
		int len = p->tot_len;
		while (len > 0) {
			if (hs->filepos < 256) {
				// copy data into header
				u8 *header = (u8 *)(&hs->dolhdr);
				int tmp_len = len;
				if (hs->filepos + tmp_len > 256)
					tmp_len = 256 - hs->filepos;
				memcpy(header+hs->filepos, data, tmp_len);
				hs->filepos+=tmp_len;
				if (hs->filepos == 256) {
					// check for elf
					if ( valid_elf_image((u32)header) ) {
						hs->type = TYPE_ELF;
						hs->filesize = 0; // no decoding
						memcpy( ELF_FILE_BUFFER, header, 256);
					} else {
						int i;
						// calculate dol size
						hs->type = TYPE_DOL;
						hs->filesize = 256;
						for (i=0; i<18; i++) {
							u32 sec_end = hs->dolhdr.sec_pos[i] + hs->dolhdr.sec_size[i];
							if (sec_end > hs->filesize) {
								hs->filesize = sec_end;
							}
						}
					}
				}
				data+=tmp_len;
				len-=tmp_len;
			} else {
				switch (hs->type) {
					case TYPE_DOL: {
						// write dol into memory
						int i, found=0;
						for (i=0; i<18 && !found ; i++) {
							if ((hs->dolhdr.sec_pos[i] <= hs->filepos) && (hs->dolhdr.sec_pos[i]+hs->dolhdr.sec_size[i] > hs->filepos)) {
								u32 tmp_offset = hs->filepos - hs->dolhdr.sec_pos[i];
								u32 tmp_len = len;
								if ( tmp_offset + tmp_len > hs->dolhdr.sec_size[i] ) {
									tmp_len = hs->dolhdr.sec_size[i] - tmp_offset;
								}
								memcpy((void*)(hs->dolhdr.sec_address[i] + tmp_offset), data, tmp_len);
								flush_code((void*)(hs->dolhdr.sec_address[i] + tmp_offset), tmp_len);
								data+=tmp_len;
								len-=tmp_len;
								hs->filepos+=tmp_len;
								found=1;
							}
						}
						if (!found) {
							// if the filepos was not found in the dol header, advance the file pointer by one byte. there could be some
							// alignment between sections, for example for DVD reads
							data++;
							len--;
							hs->filepos++;
						}
						break;
					}
					default: {
						// copy others into buffer
						memcpy(ELF_FILE_BUFFER + hs->filepos, data, len);
						hs->filepos += len;
						data += len;
						len = 0;
						break;
					}
				}
			}
		}
	}
	
	if (p != NULL) {
		pbuf_free(p);
	}
	
	if (err == ERR_OK) {
		if ((hs->filepos >= hs->filesize && hs->filesize > 0) || (hs->type == TYPE_ELF && p == NULL)) {
			printf("\n");
			close_conn(pcb, hs);
		} else {
			printf(".");
		}
	}
	
	return ERR_OK;
}
/*-----------------------------------------------------------------------------------*/
static err_t
dolup_accept(void *arg, struct tcp_pcb *pcb, err_t err)
{
	struct dolup_state *hs;
	
	tcp_setprio(pcb, TCP_PRIO_MIN);
	
	/* Allocate memory for the structure that holds the state of the connection. */
	hs = mem_malloc(sizeof(struct dolup_state));
	
	if (hs == NULL) {
		printf("dolup_accept: Out of memory\n");
		return ERR_MEM;
	}
	
	printf("dolup: accept\n");
	hs->filepos = 0;
	hs->filesize = 0;
	hs->type = TYPE_UNKNOWN;
	
	/* Tell TCP that this is the structure we wish to be passed for our
		callbacks. */
	tcp_arg(pcb, hs);
	
	/* Tell TCP that we wish to be informed of incoming data by a call
		to the dolup_recv() function. */
	tcp_recv(pcb, dolup_recv);
	
	tcp_err(pcb, conn_err);
	
	tcp_poll(pcb, dolup_poll, 4);
	return ERR_OK;
}
/*-----------------------------------------------------------------------------------*/
void
dolup_init(void)
{
	struct tcp_pcb *pcb;
	
	pcb = tcp_new();
	tcp_bind(pcb, IP_ADDR_ANY, 4000);
	pcb = tcp_listen(pcb);
	tcp_accept(pcb, dolup_accept);
}
