#include <dvd.h>
#include <console.h>
#include <linux/string.h>
#include <pad.h>
#include <processor.h>
#include <cache.h>
#include "lowmem.h"
#include "ipl.h"

typedef int my_dvd_read(void *, int, int);

static char *ccodes[]={"JPN", "USA", "PAL", "???"};

void load_app(my_dvd_read *dvd_read)
{
	char *buffer = (char*)0xC0100000;
	void (*app_init)(void (*report)(const char *fmt, ...));
	int (*app_main)(void **dst, int *size, int *offset);
	void *(*app_final)();
	void (*entrypoint)();
	
	void (*app_entry)(void(**init)(const char *fmt, ...), int (**main)(), void *(**final)());

    /* now we initialize the dvdrom. this will read the disc id. */
	printf("init dvd..\n");
	if (dvd_init())
	{
		printf("dvd init failed (%s)!\n", dvd_get_error_message());
		return;
	}
	
	ipl_set_config(2); /* disable modified bios */

		/* we need to disable external IRQs */
	exception_close();

	dvd_set_streaming();
	
		/* we check if we have a fitting video mode selected, otherwise the */
		/* game *will* OSFatal because it can't change PAL <-> NTSC. */
		/* we ignore the fact that there are different hardwares out there */
		/* and hope the user's tvset is able to cope with that. good luck. */
	printf("checking countrycode...");
	dvd_read(buffer, 0x20, 0x440);
	int ccode = *(int*)(buffer + 0x18);
	if (ccode >= 3)
		ccode = 3;
	printf(" found %s.\n", ccodes[ccode]);
	
	int current_mode = lowmem_current_video_mode & 1; /* 0, 2, 4 are NTSC-styled modes, 1, 3, 5 are PAL-styled modes. */
	int required_mode = (ccode == 2) ? 1 : 0; /* pal for PAL, otherwise NTSC */
	if (current_mode != required_mode)
	{
		printf("adjusting video mode to %s\n", required_mode ? "PAL" : "NTSC");
		lowmem_current_video_mode = required_mode;
	}
	
	video_init((*(unsigned long*)0x800000cc) & 1);

		/* load the apploader-header. */
	printf("reading apploader!\n");
	memset(buffer, 0, 0x20);
	printf("dvd_read\n");
	dvd_read(buffer, 0x20, 0x2440);
	printf("apploader date: %s\n", buffer);
	printf("apploader entry point: %08lx\n", *(unsigned long*)(buffer+0x10));
	printf("apploader size: %08lx\n", *(unsigned long*)(buffer+0x14));
	
		/* read the apploader */
	dvd_read((void*)0x81200000, ((*(unsigned long*)(buffer+0x14)) + 31) &~31, 0x2460);
	
		/* call abbloader entry */
	app_entry = *(unsigned long*)(buffer+0x10);
	app_entry(&app_init, &app_main, &app_final);
	
		/* intialize it with our debug printf */
	app_init(printf);
	
		/* call the apploader until it's done. */
	while (1)
	{
		void *dst=0;
		int len=0, offset=0;
		printf(".");
		int res = app_main(&dst, &len, &offset);
		if (!res)
			break;
		if (dvd_read(dst, len, offset))
			printf("DVD READ ERROR: %s\n", dvd_get_error_message());

		flush_code(dst, len);
	}

	entrypoint = app_final();

	entrypoint();
	while (1);
}
