#include <irq.h>

struct irq_handler_s
{
	irq_handler_proc *handler;
	void *context;
};

static struct irq_handler_s irq_handler[32];

void irq_ext()
{
	int i;
	unsigned long cause = *(volatile unsigned long*)0xCC003000;
	cause &= *(volatile unsigned long*)0xCC003004;
	for (i = 0; i < 32; ++i)
		if (cause & (0x80000000>>i))
			if (irq_handler[i].handler)
				irq_handler[i].handler(i, irq_handler[i].context);
	*(volatile unsigned long*)0xCC003000 = cause;
}

int request_irq(int irq, irq_handler_proc *handler, void *context)
{
	if (irq_handler[irq].handler)
		return -1;
	irq_handler[irq].handler = handler;
	irq_handler[irq].context = context;
	*(unsigned long*)0xCC003004 |= 0x80000000>>irq;  // enable irq
	return 0;
}

int free_irq(int irq, void *context)
{
	if (!irq_handler[irq].handler)
		return -1;
	if (irq_handler[irq].context != context)
		return -2;
	
	irq_handler[irq].handler = 0;
	
	*(unsigned long*)0xCC003004 &= ~(0x80000000>>irq);
	return 0;
}

