
#include "processor.h"
#include "exi.h"
#include "console.h"
#include "pad.h"
#include "video.h"
#include "color.h"
#include "time.h"
#include "sram.h"
#include "sd.h"
#include <linux/string.h>
#include <network.h>
#include <dvd.h>
#include <context.h>
#include <irq.h>
#include <exception.h>
#include <cache.h>

#include "ipl.h"

#include "lwip/udp.h"
#include "lwip/ip_addr.h"

#include "gdbstub/gdbstub.h"

typedef int my_dvd_read(void *, int, unsigned int);

void load_app(my_dvd_read *dvd);
extern void _start(), _end();
void start_dvd_serv();

void lowmem_init()
{
	int memsize = 0x1800000;
	
	// 0x80000000+n
	memset((void*)0x80000000, 0, 0x100);
	
	// 0..f  game-info
	strcpy((void*)0x80000000, "GPOP8P");        // PSO - could be anything.
	*(unsigned long*)0x80000008 = 1;
	*(unsigned long*)0x8000001c = 0xc2339f3d;   // DVD magic word
	
	*(unsigned long*)0x80000020 = 0x0d15ea5e;   // magic word "disease"
	*(unsigned long*)0x80000024 = 1;            // version
	*(unsigned long*)0x80000028 = memsize; // memory size
			// retail. 
	*(unsigned long*)0x8000002c = 1 + ((*(unsigned long*)0xCC00302c)>>28);

	*(unsigned long*)0x80000030 = 0;
	*(unsigned long*)0x80000034 = 0x816FFFF0;		// homebrew stuff get their SP from here... Usually set by the apploader
	*(unsigned long*)0x80000038 = 0;
	*(unsigned long*)0x8000003c = 0;
	
	*(unsigned long*)0x800000cc = 1; // video mode
	
	*(unsigned long*)0x800000EC = 0x80000000 | memsize; // top of memory?
	*(unsigned long*)0x800000F0 = memsize; // simulated memory size
	*(unsigned long*)0x800000F4 = 0;
	*(unsigned long*)0x800000F8 = 162000000;  // bus speed: 162 MHz
	*(unsigned long*)0x800000FC = 486000000;  // cpu speed: 486 Mhz

	*(unsigned short*)0x800030e0 = 6; // production pads

	dcache_flush((void*)0x80000000, 0x100);
}

extern int decCount;

int reset_switch(int irq, void *ctx)
{
	*(unsigned long*)0xcc003024=0;
	while (1);
}

static struct console_data_s console;

	/* the menu stuff. */

struct menuentry_s;

void menu_backup(struct menuentry_s *entry, int event, int data);
void menu_boot_ipl(struct menuentry_s *entry, int event, int data);
void menu_save(struct menuentry_s *entry, int event, int data);
void menu_ip(struct menuentry_s *entry, int event, int data);
void menu_boot_backup(struct menuentry_s *entry, int event, int data);
void menu_boot_game(struct menuentry_s *entry, int event, int data);
void menu_boot_sd(struct menuentry_s *entry, int event, int data);
void menu_debug(struct menuentry_s *entry, int event, int data);

#define EVENT_INIT   0
#define EVENT_REDRAW 1
#define EVENT_BUTTON 2

#define BUTTON_A 0
#define BUTTON_LEFT 1
#define BUTTON_RIGHT 2
#define BUTTON_UP 3
#define BUTTON_DOWN 4

int menu_hilight = -1;
int lock_menu=0;

struct menuentry_s
{
	const char *text, *value;
	void (*handler)(struct menuentry_s *, int event, int data);
	int data;
	char temp[20];
} menu[]=
	{
		{"Backup Game", 0, menu_backup},
		{"Boot original IPL", 0, menu_boot_ipl, -1},
		{"Boot Game Backup", 0, menu_boot_backup},
		{"Boot Game DVD", 0, menu_boot_game},
		{"IP Address", 0, menu_ip, 0},
		{"Subnet Mask", 0, menu_ip, 1},
		{"Gateway", 0, menu_ip, 2},
		{"Apply Settings and Reboot", 0, menu_save},
		{"Boot from SD", 0, menu_boot_sd},
		{"Debug",0,menu_debug},
		{0,0,0},
	};


void menu_backup(struct menuentry_s *entry, int event, int data)
{
	if ((event == EVENT_BUTTON) && (data == BUTTON_A)) {
	    while(1) {	    
    		struct PAD pad;
    		do { PAD_ReadState(&pad, 0); } while (pad.Digital.A); 
    	    printf("\n\n\n\n\n\n\n\n\n\n\n\n\n\n\n\n\n\n\n\n\n\n\n\n\n\n\n\n\n\n\n\n\n\n\n\n\n\n");
        	set_position(0, 50);    	
    		printf("     Stopping motor...\n");
    	    dvd_motor_off();
    		printf("     Insert original game and press A when ready...\n");
    		do { PAD_ReadState(&pad, 0); } while (!pad.Digital.A); 
    	    printf ("     Please Wait...\n");	    
    	    dvd_reset();
        	if (dvd_init())
        	{
        		printf("dvd init failed (%s)!\n", dvd_get_error_message());
        	    for(;;);
        	}
    		struct ip_addr *d = sram_get_ip(entry->data);
    		char temp[256];
    		sprintf(temp, "%d.%d.%d.%d", 
    			ip4_addr1(d), ip4_addr2(d), ip4_addr3(d), ip4_addr4(d));
        	
        	printf ("\n     Now get file http://%s/backup.iso\n", temp);
            start_dvd_serv();    	
            printf ("\n     Press A if you want to rip another disk\n");
    		do { PAD_ReadState(&pad, 0); } while (pad.Digital.A); 
    		do { PAD_ReadState(&pad, 0); } while (!pad.Digital.A); 
	    }
	}		
}

void menu_debug(struct menuentry_s *entry, int event, int data)
{
	unsigned char buffer[256];
	if ((event == EVENT_BUTTON) && (data == BUTTON_A)) {
		ipl_rob_read(buffer,0,256);
		int i;
		for (i=0;i<256;i++) printf("%02x ",buffer[i]);
		for(;;);
	}
}

void menu_boot_sd(struct menuentry_s *entry, int event, int data)
{
	if ((event == EVENT_BUTTON) && (data == BUTTON_A)) {
		printf("Booting from SD\n");

		sd_boot();
		printf("Booting from SD done\n");
		lock_menu=1;
	}
		
}

void menu_boot_ipl(struct menuentry_s *entry, int event, int data)
{
	if ((event == EVENT_BUTTON) && (data == BUTTON_A))
		ipl_boot(entry->data);
}

void menu_boot_backup(struct menuentry_s *entry, int event, int data)
{
	if ((event == EVENT_BUTTON) && (data == BUTTON_A))
		load_app(dvd_read);
}

void menu_boot_game(struct menuentry_s *entry, int event, int data)
{
	if ((event == EVENT_BUTTON) && (data == BUTTON_A))
	{
    	dvd_reset();
		load_app(dvd_read);
	}
}

void menu_save(struct menuentry_s *entry, int event, int data)
{
	if ((event == EVENT_BUTTON) && (data == BUTTON_A))
	{
		write_sram();
	  *(unsigned long*)0xcc003024=0;
		while (1);
	}
}

void menu_ip(struct menuentry_s *entry, int event, int data)
{
	switch (event)
	{
	case EVENT_REDRAW:
	{
		struct ip_addr *d = sram_get_ip(entry->data);
		sprintf(entry->temp, "%03d.%03d.%03d.%03d", 
			ip4_addr1(d), ip4_addr2(d), ip4_addr3(d), ip4_addr4(d));
		entry->value = entry->temp;
		break;
	}
	case EVENT_BUTTON:
		switch (data)
		{
		case BUTTON_A:
			if (menu_hilight != -1)
				menu_hilight = -1;
			else
				menu_hilight = 0;
			break;
		case BUTTON_RIGHT:
			if (menu_hilight == -1)
				break;
			menu_hilight++;
			if ((menu_hilight&3) == 3)
				++menu_hilight;
			if (menu_hilight == 16)
				menu_hilight = 0;
			break;
		case BUTTON_LEFT:
			if (menu_hilight == -1)
				break;
			if (!menu_hilight)
				menu_hilight = 14;
			else
			{
				menu_hilight--;
				if ((menu_hilight&3) == 3)
					--menu_hilight;
			}
			break;
		case BUTTON_UP:
		case BUTTON_DOWN:
		{
			if (menu_hilight == -1)
				break;
			int h = 14 - menu_hilight;
			struct ip_addr *d = sram_get_ip(entry->data);
			int oct = (d->addr >> (((h&~3)*2))) & 0xFF;

			int f = 1;
			int i;
			for (i=0; i<(h&3); ++i)
				f *= 10;
			int n = oct / f % 10;
			oct -= n*f;
			
			if (data == BUTTON_UP)
				n++;
			else
				n+=9;
			n%=10;
			
			oct += n*f;
			if (oct > 255)
				oct = 200;
			d->addr &= ~(0xFF << (((h&~3)*2)));
			d->addr |= oct  << (((h&~3)*2));
			break;
		}
		}
	}
}

int main()
{
	u32* fb = (u32*)0xC0F00000;
	
	// clear interrupt mask
	*(volatile unsigned long*)0xCC003004 = 0;
	
	memset((void*)0x80000000, 0, ((int)_start) - 0x80000000);
	memset(_end, 0, 0x81800000 - ((int)_end));
	dcache_flush((void*)0x80000000, 0x1800000);

//	video_init(VIDEO_640X480_PAL50_YUV16);
	video_set_framebuffer(VIDEO_FRAMEBUFFER_1, ((u32)fb));
	video_set_framebuffer(VIDEO_FRAMEBUFFER_2, ((u32)fb) + 1280);

	console_init(&console, fb, 640, 576, 640*2);

	printf("\n\n\n\nlowmem ");	
	lowmem_init();
	printf("pad ");
	pad_init();
	printf("exi ");
	exi_init();
	printf("sram ");
	read_sram();
	printf("video ");

//	int videomode = sram_get_videomode();
    int videomode = video_get_mode();
	
	if (videomode)
		video_init(VIDEO_640X480_PAL50_YUV16);
	else
		video_init(VIDEO_640X480_NTSC_YUV16);

	printf("done..\n");

	console_init(&console, fb, 640, videomode ? 576 : 480, 640*2);
	set_position(30, 50);
	printf("IPL REPLACEMENT 0.3333 (C) 2003,2004 tmbinc\n");
	set_position(30, 66);
	printf("VIPER RIPPER special version by K-0-D\n");
	

	exception_init();
	
	tb_t start, now;

	mftb(&start);
	
	mtdec(40500000);

//	printf("--- display offset: %i\n", (s8)sram_data[0x10]);
//	printf("--- language: %i\n", sram_data[0x12]);


	// setup network interface
	network_init();
	
	mtmsr(mfmsr() | 2);
	mtmsr(mfmsr() | 0x8000); // enable external (ext, dec) interrupts

#if 0
	// activate debugger interface
	gdbstub_init();
#else
	request_irq(30, reset_switch, 0);
#endif

#if 1

#if 0
	printf("PRESS\n"
	       "X for Game boot\n"
	       "Y for original IPL\n"
	       "Z for alternative BIOS (at 1MB)\n");
#endif
/*	if (ipl_rob_test())
		printf(" Rob's IPL chip found\n");
	else
		printf(" Rob's IPL chip not found\n");*/

	printf(
		" - At first run, set the IP config and Apply Settings.\n"
		" - Otherwise select what you want to boot from this menu.\n"
		" - To upload a binary using BBA, just upload it to port 4000.\n"
		"   It can either be ELF or DOL\n");

#if 0
		/* npdp stuff by Dr.Matrix. According to him, should fix something. */
	unsigned char npdp_id[0x40];
	if (!npdp_getid(npdp_id))
	{
		printf("NPDP found!\n");
		printf("%s\n", npdp_id + 0x10);
	}
#endif

	int menu_redraw = 1;
	
	int menu_current = 0;
	int waitrelease = 0;
	
	while (1) {
		struct PAD pad;
		mftb(&now);
		network_poll();
		
			/* sorry, the gx menu didn't finished in time :( */
			
			/* this code is so *damn* crappy. i'll replace it as */
			/* soon as possible, but until that, is has to stay. */
		if (menu_redraw && !lock_menu)
		{
			struct menuentry_s *m = menu;
			int x = 80;
			int y = 240;
			int i = 0;
			while (m->handler)
			{
				m->handler(m, EVENT_REDRAW, 0);
				
				if (i == menu_current)
					set_color(COLOR_WHITE, COLOR_BLACK);
				else
					set_color(COLOR_BLACK, COLOR_WHITE);
				set_position(x, y);
				printf(" %-20s", m->text);
				int a;
				int l = m->value ? strlen(m->value) : 0;
				for (a=0; a<20-l; ++a)
					putchar(' ');
				for (a=0; a<l; ++a)
				{
					int hi = (a == menu_hilight) && (i == menu_current);
					if (hi)
						set_color(COLOR_RED, COLOR_WHITE);
					putchar(m->value[a]);
					if (hi)
						set_color(COLOR_WHITE, COLOR_BLACK);
				}
				
				y += 18;
				++m; ++i;
			}
			set_color(COLOR_BLACK, COLOR_WHITE);
			menu_redraw = 0;
		}

		PAD_ReadState(&pad, 0);
		
		int isup=0, isdown=0, isleft = 0, isright = 0;;
		
		if (pad.Digital.Up || (pad.Analog.Y < -50))
			isup = 1;
		if (pad.Digital.Down || (pad.Analog.Y > 50))
			isdown = 1;

		if (pad.Digital.Left || (pad.Analog.X < -50))
			isleft = 1;
		if (pad.Digital.Right || (pad.Analog.X > 50))
			isright = 1;

		if (pad.Digital.A && (!waitrelease))
		{
			menu[menu_current].handler(menu + menu_current, EVENT_BUTTON, BUTTON_A);
			menu_redraw = 1;
			waitrelease = 1;
		}
		
		if (isleft && !waitrelease)
		{
			menu[menu_current].handler(menu + menu_current, EVENT_BUTTON, BUTTON_LEFT);
			waitrelease = 1;
			menu_redraw = 1;
		}
		
		if (isright && !waitrelease)
		{
			menu[menu_current].handler(menu + menu_current, EVENT_BUTTON, BUTTON_RIGHT);
			waitrelease = 1;
			menu_redraw = 1;
		}
		
		if (isup && !waitrelease)
		{
			if (menu_hilight == -1)
			{
				menu_current--;
				if (menu_current < 0)
				{
					menu_current = 0;
					while (menu[menu_current + 1].handler)
						++menu_current;
				}
			} else
				menu[menu_current].handler(menu + menu_current, EVENT_BUTTON, BUTTON_UP);
				
			menu_redraw = 1;
			waitrelease = 1;
		}
		if (isdown && !waitrelease)
		{
			if (menu_hilight == -1)
			{
				menu_current++;
				if (!menu[menu_current].handler)
					menu_current = 0;
			} else
				menu[menu_current].handler(menu + menu_current, EVENT_BUTTON, BUTTON_DOWN);
				
			menu_redraw = 1;
			waitrelease = 1;
		}
		if (!(isup || isdown || pad.Digital.A || isleft || isright))
			waitrelease = 0;

#if 0
			if (pad.Digital.X)
				load_app(dvd_read);
			if (pad.Digital.Y)
				ipl_boot(-1);
			if (pad.Digital.Z)
				ipl_boot(8);
#endif
	}
#endif
	while (1);
	return 0;
}
